// Copyright (C) 2019  Wilko Manger
// Copyright (C) 2019  Mathieu Velten
//
// This file is part of matrix-dart-sdk.
//
// matrix-dart-sdk is free software: you can redistribute it and/or modify
// it under the terms of the GNU Affero General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// matrix-dart-sdk is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Affero General Public License for more details.
//
// You should have received a copy of the GNU Affero General Public License
// along with matrix-dart-sdk.  If not, see <https://www.gnu.org/licenses/>.

import 'package:kiwi/kiwi.dart';
import 'package:matrix_sdk/matrix_sdk.dart';
import 'package:matrix_sdk/src/event/event.dart';
import 'package:matrix_sdk/src/event/room/room_event.dart';
import 'package:matrix_sdk/src/event/room/state/room_upgrade_event.dart';
import 'package:matrix_sdk/src/room/room.dart';
import 'package:matrix_sdk/src/storables.dart';
import 'package:matrix_sdk/src/identifier.dart';
import 'package:meta/meta.dart';

/// Events map that will lazily load from the given
/// [store].
class Timeline extends Storables<EventId, RoomEvent> {

  final Room _room;
  final Rooms _rooms;

  Timeline(this._room, String scope)
    : _rooms = Container().resolve(scope), super(scope);

  @override
  @protected
  Future<RoomEvent> getSingleFromStore(EventId id)
    => store?.getRoomEvent(_room, id);

  @override
  @protected
  Stream<RoomEvent> getFromStore({int upTo, RoomEvent until})
    => store?.getRoomEvents(
      _room,
      count: upTo,
      until: until
    ) ?? Stream.empty();

  @override
  @protected
  Future<void> setInStore(Iterable<RoomEvent> events)
    => store?.addRoomEvents(_room, events);

  @protected
  @override
  Future<void> removeInStore(EventId id) async {
    await store?.removeEvent(id);
  }

  @protected
  @override
  Stream<RoomEvent> getFromRemote({int upTo})
    => _room.retrieveRemoteEvents(upTo);

  @override
  Iterable<RoomEvent> sort(Iterable<Event> items) {
    final list = items.toList(growable: false);
    list.sort((a, b) {
      if (a is RoomEvent && b is RoomEvent) {
        return a.time.compareTo(b.time);
      } else {
        return a.id.toString().compareTo(b.id.toString());
      }
    });

    return list.reversed;
  }

  @override
  Future<void> addAll(Iterable<RoomEvent> values) async {

    // Remove echo events
    for (final event in values) {
      final transactionId = event?.unsigned?.transactionId;
      if (transactionId != null) {
        // delete local echo
        await remove(EventId(transactionId));
      }
    }

    await super.addAll(values);
  }

  @override
  Stream<RoomEvent> get({
    int upTo,
    RoomEvent until,
    bool allowRemote = true,
    bool allowFromPreviousRoom = true}) async* {

    if (!allowFromPreviousRoom || !_room.isReplacement) {
      yield* super.get(
        upTo: upTo,
        until: until,
        allowRemote: allowRemote
      );
    } else {
      var passedCreation = false;
      var i = 0;
      await for (RoomEvent event in super.get(
          upTo: upTo,
          until: until,
          allowRemote: allowRemote
      )) {
        yield event;
        i++;

        if (event is RoomCreationEvent) {
          passedCreation = true;
        }
      }

      // If this room replaces another room, also yield
      // the events from that room
      if (passedCreation) {
        // Verify from the other room that this is indeed
        // the replacement room.
        final oldRoom = await _rooms[_room.replacesId];
        if (oldRoom.replacementId != _room.id) {
          return;
        }

        var started = false;
        await for (final event in oldRoom.timeline.get(
          upTo: upTo - i,
          until: until,
          allowRemote: allowRemote
        )) {
          if (event is RoomUpgradeEvent) {
            started = true;
          }

          if (started) {
            yield event;
          }

        }
      }
    }
  }
}
