// Copyright (C) 2019  Wilko Manger
// Copyright (C) 2019  Mathieu Velten
//
// This file is part of matrix-dart-sdk.
//
// matrix-dart-sdk is free software: you can redistribute it and/or modify
// it under the terms of the GNU Affero General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// matrix-dart-sdk is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Affero General Public License for more details.
//
// You should have received a copy of the GNU Affero General Public License
// along with matrix-dart-sdk.  If not, see <https://www.gnu.org/licenses/>.

import 'dart:async';
import 'dart:collection';

import 'package:matrix_sdk/src/storables.dart';
import 'package:matrix_sdk/src/identifier.dart';
import 'package:matrix_sdk/src/user.dart';
import 'package:meta/meta.dart';

class Members extends Storables<UserId, User> {
  final _memory = SplayTreeSet<User>();
  final RoomId _roomId;

  Members(this._roomId, String scope) : super(scope);

  @override
  @protected
  FutureOr<Iterable<User>> getFromMemory() async {
    final uniqueUsers = SplayTreeSet<User>.of(
        _memory,
            (a, b) => a.id.toString().compareTo(b.id.toString())
    );
    final memory = Future.wait(
        uniqueUsers
            .map((user) => getStateOf(user.id))
            .map((user) async => user is Future ? await user : user)
    );

    return memory;
  }

  @override
  @protected
  void setInMemory(User item) => _memory.add(item);

  @override
  @protected
  void setInMemoryIfAbsent(User item, User Function() ifAbsent) {
    final added = _memory.add(item);
    if (added) {
      ifAbsent();
    }
  }

  @override
  @protected
  void removeInMemory(UserId id)
    => _memory.removeWhere((user) => user.id == id);

  @override
  @protected
  Future<User> getSingleFromStore(UserId id) => store?.getMember(_roomId, id);

  @override
  @protected
  Future<void> setInStore(Iterable<User> users)
    => store?.addMembers(_roomId, users);

  @override
  @protected
  Stream<User> getFromStore({int upTo, User until})
    => store?.getMembers(_roomId) ?? Stream.empty();

  @override
  Future<void> removeInStore(UserId id) => null;

  @override
  Stream<User> getFromRemote({int upTo}) => Stream.empty();

  @override
  FutureOr<User> operator [](UserId id) => getStateOf(id);

  /// Returns a [User] with the given [id]. If [at] is provided,
  /// returns a [User] with a `state` that was set at that time.
  ///
  /// If [at] is `null` (default), the most recent state is returned.
  FutureOr<User> getStateOf(UserId id, {DateTime at}) {
    if (at == null) {
      at = DateTime.now();
    }

    final users = _memory
        .where((user) => user.id == id)
        .toList(growable: false);

    // If even the first user state is before `at`,
    // return that one.
    if (users.isNotEmpty) {
      final user = users.first;

      if (at.millisecondsSinceEpoch < user.state.since.millisecondsSinceEpoch) {
        return user;
      }
    }

    for (var i = 0; i < users.length; i++) {
      final current = users[i];
      final next = i + 1 < users.length ? users[i + 1] : null;

      final currentIsBefore =
        current.state.since.millisecondsSinceEpoch <= at.millisecondsSinceEpoch;

      final nextIsAfterOrNull = next == null
        || next.state.since.millisecondsSinceEpoch > at.millisecondsSinceEpoch;

      if (currentIsBefore && nextIsAfterOrNull) {
        return current;
      }
    }

    // If not in memory, return from store
    return store?.getMember(_roomId, id, at: at);
  }
}
