// Copyright (C) 2019  Wilko Manger
// Copyright (C) 2019  Mathieu Velten
//
// This file is part of matrix-dart-sdk.
//
// matrix-dart-sdk is free software: you can redistribute it and/or modify
// it under the terms of the GNU Affero General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// matrix-dart-sdk is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Affero General Public License for more details.
//
// You should have received a copy of the GNU Affero General Public License
// along with matrix-dart-sdk.  If not, see <https://www.gnu.org/licenses/>.

import 'dart:async';

import 'package:kiwi/kiwi.dart';
import 'package:matrix_sdk/src/identifier.dart';
import 'package:matrix_sdk/src/store.dart';
import 'package:meta/meta.dart';

/// [Identifiable] objects that can possibly be stored.
abstract class Storables<I extends Id, T extends Identifiable<I>> {
  final Map<I, T> _memory = Map();

  final Store store;

  Storables(String scope) : store = Container().resolve(scope);

  @protected
  FutureOr<Iterable<T>> getFromMemory() => _memory.values;

  @protected
  void setInMemory(T item) => _memory[item.id] = item;

  @protected
  void setInMemoryIfAbsent(T item, T ifAbsent()) {
    _memory.putIfAbsent(item.id, ifAbsent);
  }

  @protected
  void removeInMemory(I id) {
    _memory.remove(id);
  }

  @protected
  Future<T> getSingleFromStore(I id);

  @protected
  Stream<T> getFromStore({int upTo, T until});

  @protected
  Future<void> setInStore(Iterable<T> value);

  @protected
  Future<void> removeInStore(I id) => null;

  @protected
  Stream<T> getFromRemote({int upTo}) => Stream.empty();

  @protected
  Iterable<T> sort(Iterable<T> items) => items;

  FutureOr<T> operator [](I id) {
    // If the room is in memory, return that one
    final fromMemory = _memory[id];
    if (fromMemory != null) {
      return fromMemory;
    }

    // Look in the store
    return getSingleFromStore(id);
  }

  /// Returns up to [upTo], and/or until [until]. First from memory,
  /// then from the store. If not enough items have been yielded,
  /// looks up more items with [getFromRemote].
  ///
  /// If neither [upTo] nor [until] is given, will return all
  /// items from memory and store.
  ///
  /// If [upTo] is not given, items will not be fetched remotely.
  ///
  /// **WARNING:** Using [until] without [upTo] can be dangerous and
  /// easily loads all items, if [until] is not found for example.
  Stream<T> get({
    int upTo,
    T until,
    bool allowRemote = true
  }) async* {

    var i = 0;
    var item;

    final memory = sort(await getFromMemory());

    for (item in memory) {
      yield item;
      i++;

      if ((upTo != null && i >= upTo) || item == until) {
        return;
      }
    }

    await for (item in getFromStore(
      upTo: upTo != null ? upTo - i : null,
      until: until
    )) {
      bool shouldYield = false;

      // If the event was not already in memory, yield it
      setInMemoryIfAbsent(item, () {
        shouldYield = true;
        return item;
      });

      if (shouldYield) {
        yield item;
        i++;
      }

      if ((upTo != null && i >= upTo) || item == until) {
        return;
      }
    }

    if (!allowRemote || upTo == null) {
      return;
    }

    Iterable<T> remote = await getFromRemote(
      upTo: upTo - 1
    )?.toList();

    if (remote != null) {
      remote = sort(remote);
      await setInStore(remote);

      for (item in remote) {
        bool shouldYield = false;

        // If the event was not already in memory, yield it
        setInMemoryIfAbsent(item, () {
          shouldYield = true;
          return item;
        });

        if (shouldYield) {
          yield item;
          i++;
        }

        if ((upTo != null && i >= upTo) || item == until) {
          return;
        }
      }
    }
  }

  Future<void> addAll(Iterable<T> values) async {
    final nonNullValues = values.where((value) => value != null);

    // Add to memory
    for (final value in nonNullValues) {
      setInMemory(value);
    }

    await setInStore(nonNullValues);
  }

  Future<void> remove(I id) async {
    removeInMemory(id);
    await removeInStore(id);
  }
}
